/* urmpplaylist.c -- 
 * Created: Mon Feb 16 12:08:12 2009 by faith@acm.org
 * Revised: Wed Jul  1 19:14:02 2009 by faith@acm.org
 * Copyright 2009 Rickard E. Faith (faith@acm.org)
 * This program comes with ABSOLUTELY NO WARRANTY.
 * 
 * $Id$
 * 
 */

#include "urmpplaylist.h"
#include <time.h>
#include <dirent.h>

const char *urmp_playlist_list(void **cookie)
{
	DIR           *dir;
	struct dirent *dirent;

	if (*cookie) {
		dir = (DIR *)*cookie;
	} else {
		dir = *cookie = opendir(urmp_config_dir_path);
	}

	while ((dirent = readdir(dir))) {
		const char *pt = strrchr(dirent->d_name, '.');
		if (!pt || strncmp(".m3u", pt, 4)) continue;
		return ur_atom_string_n(dirent->d_name, pt - dirent->d_name);
	}
	closedir(dir);
	return NULL;
}

void urmp_playlist_write(const char *filename, ur_list_t list)
{
	const char *fullpath;
	char       path[urmp_max_pathlen+1];
	FILE       *str;
	time_t     t;
	void       *pathname;

	if (!filename || !filename[0]) return;
	if (filename[0] == '.' || filename[0] == '/') {
		fullpath = filename;
	} else {
		snprintf(path, urmp_max_pathlen,
			 "%s/%s.m3u", urmp_config_dir_path, filename);
		fullpath = path;
	}

	if (!(str = fopen(fullpath, "w"))) {
		ur_fatal_errno(__func__,
			       "Cannot open %s for write\n", fullpath);
	}
	time(&t);
	ur_trace(URMP_TRACE_PLAYLIST, __func__,
		 "Writing %24.24s\n", ctime(&t));
	fprintf(str, "# Generated by urmp %24.24s\n", ctime(&t));
	if (list) {
		while ((pathname = ur_list_pop(list))) {
			ur_trace(URMP_TRACE_PLAYLIST, __func__,
				 "Writing %s\n", (char *)pathname);
			fprintf(str, "%s\n", (const char *)pathname);
		}
	}
	fclose(str);
}

ur_list_t urmp_playlist_read(const char *filename)
{
	const char *fullpath;
	char       path[urmp_max_pathlen+1];
	FILE       *str;
	char       buf[urmp_max_pathlen+1];
	ur_list_t  list;

	if (!filename || !filename[0]) return NULL;
	if (filename[0] == '.' || filename[0] == '/') {
		fullpath = filename;
	} else {
		snprintf(path, urmp_max_pathlen,
			 "%s/%s.m3u", urmp_config_dir_path, filename);
		fullpath = path;
	}

	if (!(str = fopen(fullpath, "r"))) return NULL;

	list = ur_list_create();
	while (fgets(buf, urmp_max_pathlen, str)) {
		if (buf[0] == '#' || buf[0] == '!') continue;
		while (buf[strlen(buf)-1] == '\n'
		       || buf[strlen(buf)-1] == '\r') {
			buf[strlen(buf)-1] = '\0';
		}
		ur_list_append(list, ur_atom_string(buf));
	}
	fclose(str);
	return list;
}
